import itertools
import pathlib

from textual import on, work
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Grid
from textual.message import Message
from textual.screen import ModalScreen
from textual.suggester import Suggester
from textual.widgets import Button, DirectoryTree, Footer, Input, Label


class CurDirSuggestor(Suggester):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, case_sensitive=True, **kwargs)

    async def get_suggestion(self, value: str) -> str | None:
        as_path = pathlib.Path(value)
        if not as_path.is_dir():
            as_path = as_path.parent
        if as_path.exists():
            for x in sorted(itertools.islice(as_path.iterdir(), 100)):
                if str(x).startswith(value):
                    return str(x)

        return None


class SaveDataframe(Message):
    def __init__(self, path: pathlib.Path) -> None:
        super().__init__()
        self.path = path


class OverwriteWarning(ModalScreen):
    DEFAULT_CSS = """
OverwriteWarning {
 align: center middle;

}
    
OverwriteWarning > #dialog {
    grid-size: 2;
    grid-gutter: 1 2;
    grid-rows: 1fr 3;
    padding: 0 1;
    width: 60;
    height: 11;
    border: thick $background 80%;
    background: $surface;
}

OverwriteWarning > #dialog > #question {
    column-span: 2;
    height: 1fr;
    width: 1fr;
    content-align: center middle;
}

OverwriteWarning > #dialog > Button {
    width: 100%;
}
"""

    def __init__(self, target: pathlib.Path):
        super().__init__()
        self._target = target

    def compose(self) -> ComposeResult:
        yield Grid(
            Label(f"Overwrite existing file {self._target}?", id="question"),
            Button("Overwrite", variant="error", id="overwrite"),
            Button("Cancel", variant="primary", id="cancel"),
            id="dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "overwrite":
            self.dismiss(True)
        else:
            self.dismiss(False)


class SaveModal(ModalScreen):
    SUPPORTED_EXTS = ("arrows", "arrow", "feather", "json", "parquet", "pqt", "csv")

    DEFAULT_CSS = """
SaveModal {
 align: center middle;

}
    
SaveModal > #dialog {
    grid-size: 1 5;
    grid-gutter: 1;
    grid-rows: 1 1fr 3 3 1;
    padding: 0 1;
    width: 60;
    height: 30;
    border: $primary 80%;
    background: $surface;
}

"""

    BINDINGS = [
        ("escape", "close()", "Close"),
        Binding("tab", "toggle_tab()", show=False),
    ]

    def compose(self) -> ComposeResult:
        with Grid(id="dialog"):
            yield Label("Save dataframe to disk")
            yield DirectoryTree("./", name="Current Directory")
            yield Input(placeholder="Enter filename...", select_on_focus=False, suggester=CurDirSuggestor())
            yield Button("Save", variant="primary")
            foot = Footer()
            foot.styles.padding = 0
            yield foot

    def action_close(self):
        self.dismiss(None)

    def action_toggle_tab(self):
        box = self.query_one(Input)
        save = self.query_one(Button)
        if self.query_one(DirectoryTree).has_focus:
            box.focus()
        if box.has_focus:
            if box._suggestion:  # pylint: disable=protected-access
                box.action_cursor_right()
            else:
                save.focus()

    def on_mount(self):
        self.query_one(DirectoryTree).focus()

    @on(Input.Submitted)
    def on_enter(self):
        self.query_one(Button).focus()

    @on(DirectoryTree.FileSelected)
    def on_file_select(self):
        (inp := self.query_one(Input)).focus()
        inp.cursor_position = len(inp.value)

    @on(DirectoryTree.NodeHighlighted)
    def on_highlight(self, msg: DirectoryTree.NodeHighlighted):
        self.query_one(Input).value = str(pathlib.Path(msg.node.data.path))  # type: ignore

    @on(Button.Pressed)
    @work
    async def on_save_selected(self):
        target = self.query_one(Input).value
        if not target:
            self.notify("Must specify a filename to save to", severity="error", timeout=3)
            self.query_one(Input).focus()
            return
        target_path = pathlib.Path(target)
        if target_path.is_dir():
            self.notify(f"Must specify a filename to save to in directory {target}", severity="error", timeout=3)
            self.query_one(Input).focus()
            return

        if target_path.suffix.lstrip(".") not in SaveModal.SUPPORTED_EXTS:
            self.notify(
                f"Target filename must end with one of {'/'.join(SaveModal.SUPPORTED_EXTS)}",
                severity="error",
                timeout=3,
            )
            self.query_one(Input).focus()
            return

        if target_path.exists():
            confirm = await self.app.push_screen_wait(OverwriteWarning(target_path))
            if not confirm:
                self.query_one(Input).focus()
                return

        self.dismiss(target_path)



from __future__ import annotations

import datetime
from enum import Enum, auto
from itertools import accumulate
from typing import Any, ClassVar, cast

import polars as pl
import polars.datatypes as pld
import rich.repr
from polars.interchange.protocol import Column
from rich.align import Align
from rich.console import Console, RenderableType
from rich.errors import MarkupError
from rich.markup import escape
from rich.protocol import is_renderable
from rich.segment import Segment
from rich.style import Style
from rich.text import Text
from textual import events
from textual.coordinate import Coordinate
from textual.geometry import Region, Size
from textual.message import Message
from textual.reactive import Reactive
from textual.scroll_view import ScrollView
from textual.strip import Strip

from dt_browser import COLOR_COL, COLORS, DISPLAY_IDX_COL, INDEX_COL


def polars_list_to_string(column: pl.Expr):
    return pl.concat_str(pl.lit("["), column.cast(pl.List(pl.String)).list.join(", "), pl.lit("]")).alias(
        column.meta.output_name()
    )


def cell_formatter(obj: object, null_rep: Text, col: Column | None = None) -> RenderableType:
    """Convert a cell into a Rich renderable for display.

    For correct formatting, clients should call `locale.setlocale()` first.

    Args:
        obj: Data for a cell.
        col: Column that the cell came from (used to compute width).

    Returns:
        A renderable to be displayed which represents the data.
    """
    if obj is None:
        return Align(null_rep, align="center")
    if isinstance(obj, str):
        try:
            rich_text: Text | str = Text.from_markup(obj)
        except MarkupError:
            rich_text = escape(obj)
        return rich_text
    if isinstance(obj, bool):
        return Align(
            f"[dim]{'✓' if obj else 'X'}[/] {obj}{' ' if obj else ''}",
            style="bold" if obj else "",
            align="right",
        )
    if isinstance(obj, (float, pl.Decimal)):
        return Align(f"{obj:n}", align="right")
    if isinstance(obj, int):
        if col is not None and col.is_id:
            # no separators in ID fields
            return Align(str(obj), align="right")
        return Align(f"{obj:n}", align="right")
    if isinstance(obj, (datetime.datetime, datetime.time)):
        return Align(obj.isoformat(timespec="milliseconds").replace("+00:00", "Z"), align="right")
    if isinstance(obj, datetime.date):
        return Align(obj.isoformat(), align="right")
    if isinstance(obj, datetime.timedelta):
        return Align(str(obj), align="right")
    if not is_renderable(obj):
        return str(obj)

    return cast(RenderableType, obj)


def measure_width(obj: object, console: Console) -> int:
    renderable = cell_formatter(obj, null_rep=Text(""))
    return console.measure(renderable).maximum


HEADER_HEIGHT = 1
COL_PADDING = 1
PADDING_STR = " " * COL_PADDING


class CustomTable(ScrollView, can_focus=True, inherit_bindings=False):

    class CursorType(Enum):
        ROW = auto()
        CELL = auto()
        NONE = auto()

    class CellHighlighted(Message):
        """Posted when the cursor moves to highlight a new cell.

        This is only relevant when the `cursor_type` is `"cell"`.
        It's also posted when the cell cursor is
        re-enabled (by setting `show_cursor=True`), and when the cursor type is
        changed to `"cell"`. Can be handled using `on_data_table_cell_highlighted` in
        a subclass of `DataTable` or in a parent widget in the DOM.
        """

        def __init__(
            self,
            data_table: CustomTable,
            value: Any,
            coordinate: Coordinate,
        ) -> None:
            self.data_table = data_table
            """The data table."""
            self.value: Any = value
            """The value in the highlighted cell."""
            self.coordinate: Coordinate = coordinate
            """The coordinate of the highlighted cell."""
            super().__init__()

        def __rich_repr__(self) -> rich.repr.Result:
            yield "value", self.value
            yield "coordinate", self.coordinate

        @property
        def control(self) -> CustomTable:
            """Alias for the data table."""
            return self.data_table

    class CellSelected(Message):
        """Posted by the `DataTable` widget when a cell is selected.

        This is only relevant when the `cursor_type` is `"cell"`. Can be handled using
        `on_data_table_cell_selected` in a subclass of `DataTable` or in a parent
        widget in the DOM.
        """

        def __init__(
            self,
            data_table: CustomTable,
            value: Any,
            coordinate: Coordinate,
        ) -> None:
            self.data_table = data_table
            """The data table."""
            self.value: Any = value
            """The value in the highlighted cell."""
            self.coordinate: Coordinate = coordinate
            """The coordinate of the highlighted cell."""
            super().__init__()

        def __rich_repr__(self) -> rich.repr.Result:
            yield "value", self.value
            yield "coordinate", self.coordinate

        @property
        def control(self) -> CustomTable:
            """Alias for the data table."""
            return self.data_table

    DEFAULT_CSS = """
    CustomTable:dark {
        background: initial;
    }
    CustomTable {
        background: $surface ;
        color: $text;
    }
    CustomTable > .datatable--header {
        text-style: bold;
        background: $primary;
        color: $text;
    }
    CustomTable > .datatable--cursor {
        background: $secondary;
        color: $text;
    }
    CustomTable > .datatable--even-row {
        background: $primary-background-lighten-3  10%;
    }
    """

    BINDINGS: ClassVar[list] = []

    COMPONENT_CLASSES: ClassVar[set[str]] = {"datatable--header", "datatable--cursor", "datatable--even-row"}

    cursor_coordinate: Reactive[Coordinate] = Reactive(Coordinate(0, 0), repaint=False)

    def __init__(
        self, dt: pl.DataFrame, metadata_dt: pl.DataFrame, *args, cursor_type: CustomTable.CursorType, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._dt = dt
        self._metadata_dt = metadata_dt
        self._cursor_type = self._ori_cursor_type = cursor_type

        self._lines = list[Strip]()
        self._widths: dict[str, int] = {}
        self._cum_widths: dict[str, int] = {}
        self._formatters: dict[str, pl.Expr] = {}

        self._cell_highlight: Style | None = None
        self._header_style: Style | None = None
        self._row_col_highlight: Style | None = None

        self._header: dict[str, Segment] = {}
        self._header_pad: list[Segment] = []

        self._render_header_and_table: tuple[Strip, pl.DataFrame] | None = None

        self.set_dt(dt, metadata_dt)

    def on_focus(self, event: events.Focus):
        if self._cursor_type == CustomTable.CursorType.NONE:
            self._cursor_type = CustomTable.CursorType.ROW

    def on_blur(self, event: events.Blur):
        if self._ori_cursor_type != self._cursor_type:
            self._cursor_type = self._ori_cursor_type
            self._render_header_and_table = None

    def on_mount(self):
        self._cell_highlight = self.get_component_rich_style("datatable--cursor")
        self._header_style = self.get_component_rich_style("datatable--header")
        self._row_col_highlight = self.get_component_rich_style("datatable--even-row")
        self._header_pad = [Segment(PADDING_STR, style=self._header_style)]
        self._build_header_contents()

    def _build_header_contents(self):
        self._header: dict[str, Segment] = (
            {}
            if not self._dt.columns
            else {
                x.strip(): Segment(
                    f"{PADDING_STR}{x.rjust(self._widths[x]) if dtype.is_numeric() or dtype.is_temporal() else x.ljust(self._widths[x])}",
                    style=self._header_style,
                )
                for x, dtype in self._dt.schema.items()
            }
        )

        _, header_width = self._build_base_header(self._dt.columns)
        # Max to handle empty dataframe message
        self.virtual_size = Size(header_width, max(1, len(self._dt)) + HEADER_HEIGHT)

    def set_metadata(self, metadata_dt: pl.DataFrame):
        self._metadata_dt = metadata_dt
        self._render_header_and_table = None
        self.refresh(repaint=True)

    def set_dt(self, dt: pl.DataFrame, metadata_dt: pl.DataFrame):
        if not dt.columns:
            raise Exception("Cannot display a datatable with no columns")
        self._dt = dt
        self._metadata_dt = metadata_dt
        self._set_widths({x: max(len(x), self._measure(self._dt[x])) for x in self._dt.columns})
        self._render_header_and_table = None
        self._formatters = {x: self._build_cast_expr(x, padding=self._widths[x]) for x in self._dt.columns}
        self._build_header_contents()
        self.scroll_to(0, 0, animate=False)

    def _set_widths(self, widths: dict[str, int]):
        self._widths = widths
        self._cum_widths = {
            k: v - (self._widths[k] + COL_PADDING)
            for k, v in zip(self._dt.columns, accumulate(x + COL_PADDING for x in self._widths.values()))
        }

    def render_line(self, y, *_):
        if y >= len(self._lines):
            pad = " " * (self.content_region.width)
            return Strip([Segment(pad)])
        return self._lines[y]

    def _find_minimal_x_offset(self, coordinate: Coordinate):
        col_name = self._dt.columns[coordinate.column]
        effective_width = self.scrollable_content_region.width - COL_PADDING
        free_space = effective_width - (self._widths[col_name] + COL_PADDING)

        idx = self.cursor_coordinate.column - 1
        x_offset = self._cum_widths[col_name]
        while idx >= 0:
            test_col = self._dt.columns[idx]
            free_space -= self._widths[test_col] + COL_PADDING
            if free_space >= 0:
                idx -= 1
                x_offset = self._cum_widths[test_col]
            else:
                break
        return x_offset

    def _is_col_visible(self, column: int):
        x_offset = self.scroll_offset.x
        col_name = self._dt.columns[column]
        needed_x_offset = self._cum_widths[col_name]
        needed_max_x = needed_x_offset + self._widths[col_name]
        if x_offset >= needed_x_offset or needed_max_x >= (x_offset + self.scrollable_content_region.width):
            return False
        return True

    def _is_coordinate_visible(self, coordinate: Coordinate):
        y_offset = self.scroll_offset.y
        row_offset = coordinate.row - y_offset
        if row_offset < 0 or row_offset >= (self.window_region.height - 2):
            return False
        return self._is_col_visible(coordinate.column)

    def move_cursor(self, column: int | None, row: int | None):
        self.go_to_cell(
            Coordinate(
                row=self.cursor_coordinate.row if row is None else row,
                column=self.cursor_coordinate.column if column is None else column,
            )
        )

    def go_to_cell(self, coordinate: Coordinate):
        cur_visible = self._is_coordinate_visible(coordinate)
        if coordinate.column != self.cursor_coordinate.column:
            # Any col change requires re-gening col strings due to concat tuples changing
            self._render_header_and_table = None
        elif not cur_visible:
            # rengen if row not currently displayed
            self._render_header_and_table = None

        self.cursor_coordinate = coordinate

        # If it was off-screen, scroll to it, else refresh just to update the coloring
        if not cur_visible:
            self.app.log(
                f"Cell {coordinate} was determined to not be visible in scroll offset={self.scroll_offset}, scrolling..."
            )
            self.scroll_to(y=coordinate.row, x=self._find_minimal_x_offset(coordinate), animate=False)
        else:
            self.app.log(
                f"Currently visible, no need to scorll.  Coord={self.cursor_coordinate} isNone={self._render_header_and_table is None}"
            )
            self.refresh(repaint=True)

    def on_resize(self, event: events.Resize):
        # Check maxmimal selection of new size
        if event.size.width == 0:
            return
        self._ensure_cursor()

    def _ensure_cursor(self, allow_refresh: bool = True):
        self._render_header_and_table = None

        max_idx = self.cursor_coordinate.column
        while not self._is_col_visible(max_idx) and max_idx > 0:
            max_idx -= 1

        if max_idx != self.cursor_coordinate.column or self.scrollable_content_region.height < (
            self.cursor_coordinate.row - self.scroll_offset.y
        ):
            cur_row = self.cursor_coordinate.row
            max_row = self.scroll_offset[1] + (self.scrollable_content_region.height - 1)
            self.app.log(
                f"Going to cell due to resize.  coord={self.cursor_coordinate}, NewCol={min(max_idx, self.cursor_coordinate.column)} MaxIdx={max_idx}, height={self.scrollable_content_region.height}"
            )
            self.go_to_cell(Coordinate(row=min(cur_row, max_row), column=min(max_idx, self.cursor_coordinate.column)))
        elif allow_refresh:
            self.app.log(f"Regenerating on resize. MaxIdx={max_idx}, cursor={self.cursor_coordinate}")
            self.refresh()

    def _post_cell_event(self, event_type: type[CustomTable.CellHighlighted | CustomTable.CellSelected]):
        col_name = self._dt.columns[self.cursor_coordinate.column]
        self.post_message(
            event_type(
                self,
                self._dt[self.cursor_coordinate.row, col_name],
                self.cursor_coordinate,
            )
        )

    def watch_cursor_coordinate(self):
        self._post_cell_event(CustomTable.CellHighlighted)

    def on_key(self, event: events.Key) -> None:
        x_offset, y_offset = self.scroll_offset
        requires_prep = True
        effective_height = self.scrollable_content_region.height - HEADER_HEIGHT
        old_cursor = self.cursor_coordinate
        match event.key:
            case "down":
                self.cursor_coordinate = Coordinate(
                    min(len(self._dt) - 1, self.cursor_coordinate.row + 1), self.cursor_coordinate.column
                )
                requires_prep = False
            case "up":
                self.cursor_coordinate = Coordinate(
                    max(0, self.cursor_coordinate.row - 1), self.cursor_coordinate.column
                )
                requires_prep = False
            case "pageup":
                self.cursor_coordinate = Coordinate(
                    max(0, self.cursor_coordinate.row - effective_height), self.cursor_coordinate.column
                )
            case "pagedown":
                self.cursor_coordinate = Coordinate(
                    min(len(self._dt) - 1, self.cursor_coordinate.row + effective_height),
                    self.cursor_coordinate.column,
                )
            case "left":
                if self._cursor_type == CustomTable.CursorType.CELL:
                    self.cursor_coordinate = Coordinate(
                        self.cursor_coordinate.row, max(0, self.cursor_coordinate.column - 1)
                    )
                    if self._cum_widths[self._dt.columns[self.cursor_coordinate.column]] < x_offset:
                        x_offset = self._cum_widths[self._dt.columns[self.cursor_coordinate.column]]
            case "right":
                if self._cursor_type == CustomTable.CursorType.CELL:
                    self.cursor_coordinate = Coordinate(
                        self.cursor_coordinate.row, min(len(self._dt.columns) - 1, self.cursor_coordinate.column + 1)
                    )
                    col_name = self._dt.columns[self.cursor_coordinate.column]
                    max_offset = self._cum_widths[col_name] + self._widths[col_name]
                    effective_width = self.scrollable_content_region.width
                    if max_offset >= x_offset + effective_width:
                        x_offset = self._find_minimal_x_offset(self.cursor_coordinate)

            case "home":
                self.cursor_coordinate = Coordinate(0, 0)
                x_offset = 0
            case "end":
                self.cursor_coordinate = Coordinate(len(self._dt) - 1, 0)
                x_offset = 0
            case "enter":
                self._post_cell_event(CustomTable.CellSelected)
                event.stop()
                return
            case _:
                return

        y_offset += self.cursor_coordinate.row - old_cursor.row

        if (
            self.cursor_coordinate.row
            >= (self.scroll_offset[1] + self.scrollable_content_region.height - HEADER_HEIGHT)
            or self.cursor_coordinate.row < self.scroll_offset[1]
            or x_offset != self.scroll_offset[0]
        ):
            self._render_header_and_table = None
            self.scroll_to(y=y_offset, x=x_offset, animate=False)

        elif old_cursor != self.cursor_coordinate:
            if requires_prep:
                self._render_header_and_table = None
            self.refresh(repaint=True)
        event.stop()

    def _build_cast_expr(self, col: str, padding: int = 0):
        dtype = self._dt[col].dtype
        if dtype == pld.Categorical():
            dtype = pl.Utf8
        as_str = pl.col(col).cast(pl.Utf8).fill_null("")
        if dtype.is_numeric() or dtype.is_temporal():
            return as_str.str.pad_start(padding)
        if isinstance(dtype, (pl.List, pl.Array)):
            sel = pl.col(col)
            if isinstance(dtype, pl.Array):
                sel = sel.arr.to_list()
            as_str = polars_list_to_string(sel)

        return as_str.str.pad_end(padding)

    def _build_base_header(self, cols_to_render: list[str]):
        base_header = [v for k, v in self._header.items() if k in cols_to_render] + self._header_pad
        header_width = sum(len(x.text) for x in base_header)
        return (base_header, header_width)

    @property
    def render_header_and_table(self):
        if self._render_header_and_table is None:
            scroll_x, scroll_y = self.scroll_offset

            cols_to_render: list[str] = []
            effective_width = self.scrollable_content_region.width
            if effective_width <= 2:
                return (Strip([]), pl.DataFrame())
            truncate_last: int | None = None
            for x in self._dt.columns:
                min_offset = self._cum_widths[x] - scroll_x
                max_offset = min_offset + self._widths[x]
                if min_offset < 0:
                    continue
                max_available = effective_width - min_offset - (2 * COL_PADDING)
                if max_offset >= effective_width and max_available < 4:
                    break

                cols_to_render.append(x)
                if max_offset >= effective_width:
                    truncate_last = max_available
                    break

            if not cols_to_render:
                return (Strip([]), pl.DataFrame())

            dt_height = self.window_region.height - HEADER_HEIGHT
            base_header, header_width = self._build_base_header(cols_to_render)
            excess = self.scrollable_content_region.width - header_width
            header = Strip(base_header + (self._header_pad * (excess)))
            theo_max_offset = self._cum_widths[cols_to_render[0]] + effective_width

            rend = (
                self._dt.lazy()
                .select(cols_to_render)
                .with_columns(self._metadata_dt[INDEX_COL])
                .slice(scroll_y, dt_height)
                .with_row_index(DISPLAY_IDX_COL)
            )

            formatters = self._formatters
            if truncate_last:
                formatters = self._formatters.copy()
                old_formatter = self._build_cast_expr(cols_to_render[-1], padding=0)
                formatters[cols_to_render[-1]] = (
                    pl.when(old_formatter.str.len_chars() > truncate_last - 3)
                    .then(
                        pl.concat_str(
                            old_formatter.str.slice(0, length=truncate_last - 3),
                            pl.lit("..."),
                        )
                    )
                    .otherwise(old_formatter)
                )

            visible_cols = cols_to_render.copy()
            first_col_prefix_padding = COL_PADDING
            needed_padding = theo_max_offset - self._cum_widths[visible_cols[-1]] - first_col_prefix_padding

            if COLOR_COL in self._metadata_dt.columns:
                row_colors = self._metadata_dt.slice(scroll_y, dt_height).select(
                    pl.col(COLOR_COL).cast(COLORS).alias(COLOR_COL)
                )
                cols_to_render.insert(0, COLOR_COL)
                rend = rend.with_columns(row_colors)

            def build_selector(cols: list[str], needed_padding: int = 0):
                if not cols:
                    return pl.lit("").str.pad_end(needed_padding)

                fmts = [formatters[x] for x in cols]
                if needed_padding:
                    fmts[-1] = fmts[-1].str.pad_end(needed_padding - COL_PADDING, fill_char=PADDING_STR[0])
                concat = pl.concat_str(fmts, separator=PADDING_STR, ignore_nulls=True)

                return pl.concat_str(concat, pl.lit(PADDING_STR))

            if self._cursor_type in (CustomTable.CursorType.ROW, CustomTable.CursorType.NONE):
                rend = rend.select(
                    pl.col(DISPLAY_IDX_COL),
                    pl.col(INDEX_COL),
                    (
                        pl.col(COLOR_COL)
                        if COLOR_COL in cols_to_render
                        else pl.repeat(None, pl.len(), dtype=pl.Null).alias(COLOR_COL)
                    ),
                    before_selected=build_selector(visible_cols, needed_padding),
                    selected=pl.lit(""),
                    after_selected=pl.lit(""),
                )
            else:

                cursor_col_idx = self.cursor_coordinate.column - self._dt.columns.index(visible_cols[0])

                cols_before_selected: list[str] = visible_cols[0:cursor_col_idx]
                sel_col = visible_cols[cursor_col_idx]
                cols_after_selected = visible_cols[cursor_col_idx + 1 :]

                rend = rend.select(
                    pl.col(DISPLAY_IDX_COL),
                    pl.col(INDEX_COL),
                    (
                        pl.col(COLOR_COL)
                        if COLOR_COL in cols_to_render
                        else pl.repeat(None, pl.len(), dtype=pl.Null).alias(COLOR_COL)
                    ),
                    before_selected=build_selector(cols_before_selected),
                    selected=build_selector([sel_col]),
                    after_selected=build_selector(cols_after_selected, needed_padding),
                )

            self._render_header_and_table = (header, rend.collect())
        return self._render_header_and_table

    def _get_row_bg_color_expr(self, cursor_row_idx: int) -> pl.Expr:

        return (
            pl.when(pl.col(DISPLAY_IDX_COL) == cursor_row_idx)
            .then(
                pl.lit(
                    self._row_col_highlight.bgcolor.name
                    if self._cursor_type == CustomTable.CursorType.CELL
                    else self._cell_highlight.bgcolor.name
                )
            )
            .otherwise(pl.lit(None))
        )

    def _get_sel_col_bg_color(self, struct: pl.Struct):
        return (
            self._row_col_highlight.bgcolor.name
            if self._cursor_type == CustomTable.CursorType.CELL
            else struct["bgcolor"]
        )

    def render_lines(self, crop: Region):
        # if self._render_header_and_table and self.scrollable_content_region.height - HEADER_HEIGHT != len(
        #     self._render_header_and_table
        # ):
        #     # if we get re-rendered before the resize event is passed to us
        #     print(f"Ensure during render")
        #     self._ensure_cursor(allow_refresh=False)
        #     self._render_header_and_table = None
        cur_header, render_df = self.render_header_and_table

        self._lines.clear()
        self._lines.append(cur_header)
        if render_df.is_empty():
            msg = "< Empty Dataframe >"
            padding = int((self.scrollable_content_region.width - len(msg)) / 2)
            msg = f"{' '*padding}{msg}{' '*padding}"
            self._lines.append(
                Strip(
                    [Segment(msg)],
                    cell_length=self.scrollable_content_region.width,
                )
            )
        else:
            rend = render_df.lazy()
            if self._cursor_type == CustomTable.CursorType.NONE:
                rend = rend.select(
                    segments=pl.col("before_selected").map_elements(
                        lambda x: [Segment(PADDING_STR), Segment(x)],
                        return_dtype=pl.Object,
                    )
                )
            else:
                _, scroll_y = self.scroll_offset
                cursor_row_idx = self.cursor_coordinate.row - scroll_y
                rend = rend.select(
                    segments=pl.struct(
                        pl.col("*"),
                        self._get_row_bg_color_expr(cursor_row_idx).alias("bgcolor"),
                    ).map_elements(
                        lambda struct: [
                            Segment(
                                PADDING_STR,
                                style=Style(
                                    color=(
                                        self._cell_highlight.color.name
                                        if (
                                            self._cursor_type == CustomTable.CursorType.ROW
                                            and struct[DISPLAY_IDX_COL] == cursor_row_idx
                                        )
                                        else None
                                    ),
                                    bgcolor=struct["bgcolor"],
                                ),
                            ),
                            Segment(
                                struct["before_selected"],
                                style=Style(
                                    color=(
                                        self._cell_highlight.color.name
                                        if (
                                            self._cursor_type == CustomTable.CursorType.ROW
                                            and struct[DISPLAY_IDX_COL] == cursor_row_idx
                                        )
                                        else struct[COLOR_COL]
                                    ),
                                    bgcolor=struct["bgcolor"],
                                ),
                            ),
                            Segment(
                                struct["selected"],
                                style=Style(
                                    color=(
                                        self._cell_highlight.color.name
                                        if struct[DISPLAY_IDX_COL] == cursor_row_idx
                                        else struct[COLOR_COL]
                                    ),
                                    bgcolor=(
                                        self._cell_highlight.bgcolor.name
                                        if struct[DISPLAY_IDX_COL] == cursor_row_idx
                                        else self._get_sel_col_bg_color(struct)
                                    ),
                                ),
                            ),
                            Segment(
                                struct["after_selected"],
                                style=Style(
                                    color=(
                                        self._cell_highlight.color.name
                                        if (
                                            self._cursor_type == CustomTable.CursorType.ROW
                                            and struct[DISPLAY_IDX_COL] == cursor_row_idx
                                        )
                                        else struct[COLOR_COL]
                                    ),
                                    bgcolor=struct["bgcolor"],
                                ),
                            ),
                        ],
                        return_dtype=pl.Object,
                    )
                )

            self._lines.extend(
                Strip(x, cell_length=self.scrollable_content_region.width) for x in rend.collect()["segments"]
            )

        return super().render_lines(crop)

    @staticmethod
    def can_draw(arr: pl.Series) -> bool:
        if arr.is_empty():
            return True
        dtype = arr.dtype
        if dtype == pld.Categorical():
            return CustomTable.can_draw(arr.cat.get_categories())
        if dtype.is_temporal() or dtype.is_numeric() or dtype.is_(pld.Boolean()) or dtype.is_(pld.Utf8()):
            return True
        if isinstance(dtype, (pl.List, pl.Array)):
            return True
        try:
            # try to cast
            arr.filter(arr.is_not_null()).head(10).cast(pl.Utf8())
            return True
        except pl.exceptions.PolarsError:
            return False

    def _measure(self, arr: pl.Series) -> int:
        # with some types we can measure the width more efficiently
        dtype = arr.dtype
        if dtype == pld.Categorical():
            if arr.cat.get_categories().is_empty():
                return len("<null>")
            return self._measure(arr.cat.get_categories())
        if arr.is_empty():
            return 0

        if isinstance(dtype, pl.Array):
            base = arr.arr.eval(((pl.element().cast(pl.String).str.len_chars())))
            return (base.arr.sum() + (base.arr.len() - 1) * len(", ") + 2).max()

        if isinstance(dtype, pl.List):
            base = arr.list.eval(((pl.element().cast(pl.String).str.len_chars())))
            return (base.list.sum() + (base.list.len() - 1) * len(", ") + 2).max()

        if dtype.is_integer():
            col_max = arr.max()
            col_min = arr.min()
            return max(measure_width(el, self._console) for el in [col_max, col_min])
        if dtype.is_temporal():
            try:
                value = arr.drop_nulls().slice(0, 1).cast(pl.Utf8)[0]
            except IndexError:
                return 0
            return measure_width(value, self._console)
        if dtype.is_(pld.Boolean()):
            return 7

        if arr.is_empty():
            return 0
        # for everything else, we need to compute it
        width = arr.cast(pl.Utf8(), strict=False).fill_null("<null>").str.len_chars().max()
        assert isinstance(width, int)
        return width



import datetime
import gc
import pathlib
import time
from typing import ClassVar

import click
import polars as pl
import tzlocal
from rich.spinner import Spinner
from rich.style import Style
from textual import events, on, work
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.cache import LRUCache
from textual.containers import Horizontal
from textual.reactive import reactive
from textual.timer import Timer
from textual.widget import Widget
from textual.widgets import Footer, Label, Static

from dt_browser import (
    COLOR_COL,
    COLORS,
    INDEX_COL,
    ReactiveLabel,
    ReceivesTableSelect,
    SelectFromTable,
)
from dt_browser.bookmarks import Bookmarks
from dt_browser.column_selector import ColumnSelector
from dt_browser.custom_table import CustomTable, polars_list_to_string
from dt_browser.filter_box import FilterBox
from dt_browser.save_df_modal import SaveModal
from dt_browser.suggestor import ColumnNameSuggestor

_SHOW_COLUMNS_ID = "showColumns"
_COLOR_COLUMNS_ID = "colorColumns"
_TS_COLUMNS_ID = "tsColumns"

_TIMEZONE = str(tzlocal.get_localzone())


def sizeof_fmt(num, suffix="B"):
    for unit in ("", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"):
        if abs(num) < 1024.0:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f}Yi{suffix}"


class TableWithBookmarks(CustomTable):

    DEFAULT_CSS = (
        CustomTable.DEFAULT_CSS
        + """
TableWithBookmarks > .datatable--row-bookmark {
    background: $error-lighten-3;
}

TableWithBookmarks > .datatable--row-search-result {
    background: $surface-lighten-3;
}
"""
    )

    COMPONENT_CLASSES: ClassVar[set[str]] = CustomTable.COMPONENT_CLASSES.union(
        ["datatable--row-bookmark", "datatable--row-search-result"]
    )

    active_search_queue: reactive[list[int] | None] = reactive(None)

    def __init__(self, *args, bookmarks: Bookmarks, **kwargs):
        super().__init__(*args, **kwargs)
        self._bookmarks = bookmarks
        self._bookmark_highlight: Style | None = None
        self._search_highlight: Style | None = None

    def on_mount(self):
        self._bookmark_highlight = self.get_component_rich_style("datatable--row-bookmark")
        self._search_highlight = self.get_component_rich_style("datatable--row-search-result")

    def _get_sel_col_bg_color(self, struct: pl.Struct):
        if self.active_search_queue and struct[INDEX_COL] in self.active_search_queue:
            return self._search_highlight.bgcolor.name
        if self._bookmarks.has_bookmarks and struct[INDEX_COL] in self._bookmarks.meta_dt[INDEX_COL]:
            return self._bookmark_highlight.bgcolor.name
        return super()._get_sel_col_bg_color(struct)

    def _get_row_bg_color_expr(self, cursor_row_idx: int) -> pl.Expr:
        tmp = super()._get_row_bg_color_expr(cursor_row_idx)
        if self.active_search_queue:
            tmp = (
                pl.when(pl.col(INDEX_COL).is_in(self.active_search_queue))
                .then(pl.lit(self._search_highlight.bgcolor.name))
                .otherwise(tmp)
            )
        if self._bookmarks.has_bookmarks:
            tmp = (
                pl.when(pl.col(INDEX_COL).is_in(self._bookmarks.meta_dt[INDEX_COL]))
                .then(pl.lit(self._bookmark_highlight.bgcolor.name))
                .otherwise(tmp)
            )
        return tmp


_ALREADY_DT = "dt"


def _guess_timestamp_cols(df: pl.DataFrame):
    date_range = pl.Series(values=[datetime.date(2001, 1, 1), datetime.date(2042, 1, 1)])
    converts = [(x,) + tuple(date_range.dt.epoch(x)) for x in ("s", "ms", "us", "ns")]

    for col, dtype in df.schema.items():
        if dtype.is_integer():
            for suffix, min_val, max_val in converts:
                all_in_range = (
                    df.lazy()
                    .select(is_zero=pl.col(col) == 0, is_inside=((pl.col(col) >= min_val) & (pl.col(col) <= max_val)))
                    .select(
                        count=pl.col("is_inside").any()
                        & (pl.any_horizontal(pl.col("is_zero"), pl.col("is_inside")).sum() == pl.len())
                    )
                    .collect()
                    .get_column("count")[0]
                )
                if all_in_range:
                    yield (col, suffix)
                    break
        elif dtype.is_temporal():
            yield (col, _ALREADY_DT)


class SpinnerWidget(Static):
    def __init__(self, style: str):
        super().__init__("")
        self._spinner = Spinner(style)
        self.styles.width = 1
        self.update_render: Timer | None = None

    def on_mount(self) -> None:
        self.update_render = self.set_interval(1 / 60, self.update_spinner)

    def update_spinner(self) -> None:
        self.update(self._spinner)


class TableFooter(Footer):
    DEFAULT_CSS = """
    TableFooter > .tablefooter--rowcount {
        background: $primary-darken-1;
        width: auto;
        padding: 0 2;
    }
    TableFooter > .tablefooter--search {
        background: $success-darken-1;
        width: auto;
        padding: 0 2;
    }
    TableFooter > .tablefooter--pending {
        background: $secondary-darken-1;
        width: auto;
        padding: 0 2;
    }

    FooterRowCount > .tablefooter--label {
        background: $secondary;
        text-style: bold;
    }

    """
    filter_pending = reactive(False, recompose=True)
    is_filtered = reactive(False)
    cur_row = reactive(0)
    cur_total_rows = reactive(0)
    total_rows = reactive(0)
    total_rows_display = reactive("", layout=True)

    cur_row_display = reactive(0)

    pending_action = reactive("", recompose=True)

    search_pending: reactive[bool] = reactive(False, recompose=True)
    active_search_queue: reactive[list[int] | None] = reactive(None)
    active_search_idx: reactive[int | None] = reactive(None)
    active_search_idx_display: reactive[int | None] = reactive(None)
    active_search_len: reactive[int | None] = reactive(None, recompose=True)

    def compute_active_search_len(self):
        if self.active_search_queue is None:
            return None
        return len(self.active_search_queue)

    def compute_cur_row_display(self):
        return self.cur_row + 1

    def compose(self):
        yield from super().compose()

        widths = ["auto"] * self.styles.grid_size_columns
        yield Label()
        widths.append("1fr")
        if self.pending_action:
            widths.append("auto")
            with Horizontal(classes="tablefooter--pending"):
                yield ReactiveLabel().data_bind(value=TableFooter.pending_action)
                yield Label(" ")
                yield SpinnerWidget("dots")
        if self.search_pending:
            widths.append("auto")
            with Horizontal(classes="tablefooter--search"):
                yield Label("Searching ")
                yield SpinnerWidget("dots")
        elif self.active_search_len is not None:
            widths.append("auto")
            with Horizontal(classes="tablefooter--search"):
                yield Label("Search: ")
                yield ReactiveLabel().data_bind(value=TableFooter.active_search_idx_display)
                yield Label(" / ")
                yield ReactiveLabel().data_bind(value=TableFooter.active_search_len)

        with Horizontal(classes="tablefooter--rowcount"):
            yield ReactiveLabel().data_bind(value=TableFooter.cur_row_display)
            yield Label(" / ")
            yield ReactiveLabel().data_bind(value=TableFooter.cur_total_rows)
            yield ReactiveLabel().data_bind(value=TableFooter.total_rows_display)
            if self.filter_pending:
                yield Label(" Filtering ")
                yield SpinnerWidget("dots")

        widths.append("auto")
        self.styles.grid_columns = " ".join(widths)
        self.styles.grid_size_columns = len(widths)

    def compute_total_rows_display(self):
        return f" (Filtered from {self.total_rows:,})" if self.cur_total_rows != self.total_rows else ""

    def compute_active_search_idx_display(self):
        if self.active_search_idx is None:
            return None
        return self.active_search_idx + 1


class RowDetail(Widget, can_focus=False, can_focus_children=False):

    DEFAULT_CSS = """
RowDetail {
    width: auto;
    max_width: 50%;
    min_width: 30%;
    padding: 0 1;
    border: tall $primary;
}
"""
    row_df = reactive(pl.DataFrame(), always_update=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.border_title = "Row Detail"
        self._dt = CustomTable(
            pl.DataFrame({"Temp": []}),
            pl.DataFrame().with_row_index(name=INDEX_COL).select([INDEX_COL]),
            cursor_type=CustomTable.CursorType.NONE,
        )
        self._schema: pl.DataFrame | None = None

    def on_resize(self, _: events.Resize):
        tab = self.query_one(CustomTable)
        self.can_focus_children = tab.scrollbars_enabled[0]

    def watch_row_df(self):
        if self.row_df.is_empty():
            return
        display_df = self.row_df.with_columns(
            polars_list_to_string(pl.col(x)) if isinstance(dtype, pl.List) else pl.col(x).cast(pl.Utf8)
            for x, dtype in self.row_df.schema.items()
        ).transpose(include_header=True, header_name="Field", column_names=["Value"])

        if self._schema is None:
            self._schema = pl.from_dict({k: str(v) for k, v in self.row_df.schema.items()}, strict=False).transpose(
                include_header=True, header_name="Field", column_names=["dtype"]
            )
        display_df = display_df.join(self._schema, on=["Field"]).select(["Field", "dtype", "Value"])
        coord = self._dt.cursor_coordinate

        self._dt.set_dt(display_df, display_df.with_row_index(name=INDEX_COL).select([INDEX_COL]))
        self.styles.width = self._dt.virtual_size.width + self.gutter.width + 1
        # self._dt.refresh()
        self._dt.go_to_cell(coord)

    def compose(self):
        yield self._dt


def from_file_path(path: pathlib.Path, has_header: bool = True) -> pl.DataFrame:

    if path.suffix in [".arrow", ".feather"]:
        return pl.read_ipc(path)
    if path.suffix in [".arrows", ".arrowstream"]:
        return pl.read_ipc_stream(path)
    if path.suffix == ".json":
        return pl.read_json(path)
    if path.suffix == ".csv":
        return pl.read_csv(path, has_header=has_header)
    if path.suffix == ".parquet":
        return pl.read_parquet(path)
    raise TypeError(f"Dont know how to load file type {path.suffix} for {path}")


class DtBrowser(Widget):  # pylint: disable=too-many-public-methods,too-many-instance-attributes
    """A Textual app to manage stopwatches."""

    BINDINGS = [
        ("f", "show_filter", "Filter rows"),
        ("/", "show_search", "Search"),
        ("n", "iter_search(True)", "Next"),
        Binding("N", "iter_search(False)", "Prev", key_display="shift+N"),
        ("b", "toggle_bookmark", "Add/Del Bookmark"),
        Binding("B", "show_bookmarks", "Bookmarks", key_display="shift+B"),
        Binding("c", "column_selector", "Columns..."),
        ("t", "timestamp_selector", "Timestamps..."),
        ("r", "toggle_row_detail", "Toggle Row Detail"),
        Binding("C", "show_colors", "Colors...", key_display="shift+C"),
        ("ctrl+s", "show_save", "Save dataframe as..."),
    ]

    color_by: reactive[tuple[str, ...]] = reactive(tuple(), init=False)
    visible_columns: reactive[tuple[str, ...]] = reactive(tuple())
    all_columns: reactive[tuple[str, ...]] = reactive(tuple())
    timestamp_columns: reactive[tuple[str, ...]] = reactive(tuple())
    available_timestamp_columns: reactive[tuple[str, ...]] = reactive(tuple())

    is_filtered = reactive(False)
    current_filter = reactive[str | None](None)

    cur_row = reactive(0)
    cur_total_rows = reactive(0)
    total_rows = reactive(0)

    show_row_detail = reactive(True)

    active_search_queue: reactive[list[int] | None] = reactive(None)
    active_search_idx: reactive[int | None] = reactive(None)
    active_search: reactive[str | None] = reactive(None)
    # active_dt: reactive[pl.DataFrame] = reactive(pl.DataFrame(), init=False, always_update=True)

    def __init__(self, table_name: str, source_file_or_table: pathlib.Path | pl.DataFrame):
        super().__init__()
        bt = (
            from_file_path(source_file_or_table)
            if isinstance(source_file_or_table, (str, pathlib.Path))
            else source_file_or_table
        )
        self.removed_cols = {x: v for x, v in bt.schema.items() if not CustomTable.can_draw(bt[x])}
        # bt = bt.with_columns(TestTs=datetime.datetime.now())
        self._display_dt = self._filtered_dt = self._original_dt = bt.select(
            [x for x in bt.columns if x not in self.removed_cols]
        )
        self._meta_dt = self._original_meta = self._original_dt.with_row_index(name=INDEX_COL).select([INDEX_COL])
        self._table_name = table_name
        self._bookmarks = Bookmarks()
        self._suggestor = ColumnNameSuggestor()
        self.visible_columns = tuple(self._original_dt.columns)
        self.all_columns = self.visible_columns
        self._filter_box = FilterBox(suggestor=self._suggestor, id="filter", classes="toolbox")
        self._select_interest: str | None = None
        self._column_selector = ColumnSelector(id=_SHOW_COLUMNS_ID, title="Show/Hide/Reorder Columns")
        self._color_selector = ColumnSelector(
            allow_reorder=False, id=_COLOR_COLUMNS_ID, title="Select columns to color by"
        )

        self._ts_cols = dict(_guess_timestamp_cols(self._original_dt))
        self._ts_col_selector = ColumnSelector(id=_TS_COLUMNS_ID, title="Select epoch timestamp columns")
        self._ts_col_names: dict[str, str] = {}
        self.available_timestamp_columns = tuple(self._ts_cols.keys())

        # Necessary to prevent the main table from resizing to 0 when the col selectors are mounted and then immediately resizing
        # (apparently that happens when col selector width = auto)
        self._color_selector.styles.width = 1
        self._column_selector.styles.width = 1
        self._ts_col_selector.styles.width = 1

        self._row_detail = RowDetail()

        self._color_by_cache: LRUCache[tuple[str, ...], pl.Series] = LRUCache(5)
        self._last_message_ts = time.time()

    def _set_last_message(self, *_):
        self._last_message_ts = time.time()

    def _maybe_gc(self):
        if time.time() - self._last_message_ts > 3:
            self.app.log("Triggering GC!")
            gc.collect()

    def watch_visible_columns(self):
        self._suggestor.columns = self.visible_columns

    @on(FilterBox.FilterSubmitted)
    async def update_filter(self, event: FilterBox.FilterSubmitted):
        self.current_filter = event.value
        self.apply_filter()

    @work(exclusive=True)
    async def apply_filter(self):
        if not self.current_filter:
            self.is_filtered = False
            idx = self.query_one("#main_table", CustomTable).cursor_coordinate.row
            self._set_filtered_dt(
                self._original_dt,
                self._original_meta,
                new_row=self._meta_dt[INDEX_COL][idx],
            )
        else:
            (foot := self.query_one(TableFooter)).filter_pending = True
            ctx = pl.SQLContext(frames={"dt": pl.concat([self._original_dt, self._original_meta], how="horizontal")})
            try:
                query = self.current_filter.replace(" && ", " and ").replace(" || ", " or ")
                dt = await ctx.execute(f"select * from dt where {query}").collect_async()
                meta = dt.select([x for x in dt.columns if x.startswith("__")])
                dt = dt.select([x for x in dt.columns if not x.startswith("__")])
                self.is_filtered = True
                if dt.is_empty():
                    self.notify(f"No results found for filter: {query}", severity="warning", timeout=5)
                else:
                    self._set_filtered_dt(dt, meta, new_row=0)
            except Exception as e:
                self.query_one(FilterBox).query_failed(query)
                self.notify(f"Failed to apply filter due to: {e}", severity="error", timeout=10)
                self.current_filter = None
            foot.filter_pending = False

    @on(FilterBox.GoToSubmitted)
    async def apply_search(self, event: FilterBox.GoToSubmitted):
        self.active_search = event.value

    @work(exclusive=True)
    async def watch_active_search(self, goto: bool = True):
        if not self.active_search:
            self.active_search_queue = None
            self.active_search_idx = 0
            return

        (foot := self.query_one(TableFooter)).search_pending = True
        try:
            idx_name = "__search_idx"
            ctx = pl.SQLContext(frames={"dt": self._display_dt.with_row_index(idx_name)})
            query = self.active_search.replace(" && ", " and ").replace(" || ", " or ")
            search_queue = list(
                (await ctx.execute(f"select {idx_name} from dt where {query}").collect_async())[idx_name]
            )

            foot.search_pending = False
            if not search_queue:
                self.notify("No results found for search", severity="warning", timeout=5)
            else:
                self.active_search_queue = search_queue
                self.active_search_idx = -1
                if goto:
                    # Find the nearest index to the current cursor
                    coord = self.query_one("#main_table", CustomTable).cursor_coordinate.row
                    next_row = next((i for i, x in enumerate(self.active_search_queue) if x > coord), None)
                    self.active_search_idx = next_row - 1
                    self.action_iter_search(True)
        except Exception as e:
            self.query_one(FilterBox).query_failed(query)
            self.notify(f"Failed to run search due to: {e}  {repr(e)}", severity="error", timeout=10)
            foot.search_pending = False

    def action_iter_search(self, forward: bool):
        table = self.query_one("#main_table", CustomTable)
        coord = table.cursor_coordinate
        self.active_search_idx = min(
            max(self.active_search_idx + (1 if forward else -1), 0), len(self.active_search_queue) - 1
        )
        if self.active_search_idx >= 0 and self.active_search_idx < len(self.active_search_queue):
            next_idex = self.active_search_queue[self.active_search_idx]
            ys = next_idex
            xs = table.scroll_x
            table.scroll_to(xs, ys, animate=False, force=True)
            table.move_cursor(column=coord.column, row=next_idex)
        self.refresh_bindings()

    def action_toggle_bookmark(self):
        row_idx = self.query_one("#main_table", CustomTable).cursor_coordinate.row
        did_add = self._bookmarks.toggle_bookmark(self._display_dt[row_idx], self._meta_dt[row_idx])
        self.refresh_bindings()
        self.query_one("#main_table", CustomTable).refresh(repaint=True)

        self.notify("Bookmark added!" if did_add else "Bookmark removed", severity="information", timeout=3)

    async def action_toggle_row_detail(self):
        self.show_row_detail = not self.show_row_detail

    @work
    async def action_show_save(self):
        target = await self.app.push_screen_wait(SaveModal())
        if target is None:
            return
        self.notify(f"Saving dataframe to {target}", severity="information", timeout=3)

        target = str(target)
        num_rows = len(self._display_dt)
        try:
            if target.endswith(".arrows"):
                self._display_dt.write_ipc_stream(target)
            elif target.endswith(".arrow") or target.endswith(".feather"):
                self._display_dt.write_ipc(target)
            elif target.endswith(".parquet") or target.endswith(".pqt"):
                self._display_dt.write_parquet(target)
            elif target.endswith(".json"):
                self._display_dt.write_json(target)
            elif target.endswith(".csv"):
                self._display_dt.write_csv(target)
            else:
                self.notify(f"Dont know how to write file {target}", severity="error", timeout=5)
                return
        except Exception as e:
            self.notify(f"Failed to save to {target} due to: {e}", severity="error", timeout=10)
            return

        size = pathlib.Path(target).stat().st_size
        self.notify(
            f"Successfully wrote {num_rows:,} rows / {sizeof_fmt(size)} to {target}", severity="information", timeout=5
        )

    async def watch_show_row_detail(self):
        if not self.show_row_detail:
            if existing := self.query(RowDetail):
                existing.remove()
        elif not self._display_dt.is_empty():
            await self.query_one("#main_hori", Horizontal).mount(self._row_detail)

    async def action_show_bookmarks(self):
        await self.mount(self._bookmarks, before=self.query_one(TableFooter))

    async def action_column_selector(self):
        await self.query_one("#main_hori", Horizontal).mount(self._column_selector)

    async def action_show_colors(self):

        await self.query_one("#main_hori", Horizontal).mount(self._color_selector)

    async def action_timestamp_selector(self):
        await self.query_one("#main_hori", Horizontal).mount(self._ts_col_selector)

    def _set_filtered_dt(self, filtered_dt: pl.DataFrame, filtered_meta: pl.DataFrame, **kwargs):
        self._filtered_dt = filtered_dt
        self._meta_dt = filtered_meta
        self._set_active_dt(self._filtered_dt, **kwargs)

    def _set_active_dt(self, active_dt: pl.DataFrame, new_row: int | None = None):
        ordered_cols: list[pl.Expr] = []
        for col in self.visible_columns:
            ordered_cols.append(pl.col(col))
            if col in self._ts_col_names:
                ordered_cols.append(pl.col(self._ts_col_names[col]))
        active_dt = active_dt.select(ordered_cols)

        self._display_dt = active_dt
        self.cur_total_rows = len(self._display_dt)
        if self._display_dt.is_empty():
            self.show_row_detail = False
        self.watch_active_search(goto=False)
        (table := self.query_one("#main_table", CustomTable)).set_dt(self._display_dt, self._meta_dt)
        if new_row is not None:
            table.move_cursor(row=new_row, column=None)
            self.cur_row = new_row

    @on(ColumnSelector.ColumnSelectionChanged, f"#{_SHOW_COLUMNS_ID}")
    def reorder_columns(self, event: ColumnSelector.ColumnSelectionChanged):
        self.visible_columns = tuple(event.selected_columns)
        self._set_active_dt(self._filtered_dt)

    @on(ColumnSelector.ColumnSelectionChanged, f"#{_COLOR_COLUMNS_ID}")
    async def set_color_by(self, event: ColumnSelector.ColumnSelectionChanged):
        self.color_by = tuple(event.selected_columns)

    @on(ColumnSelector.ColumnSelectionChanged, f"#{_TS_COLUMNS_ID}")
    async def set_timestamp_cols(self, event: ColumnSelector.ColumnSelectionChanged):
        self.timestamp_columns = tuple(event.selected_columns)

    @work(exclusive=True)
    async def watch_timestamp_columns(self):
        old_cols = [v for k, v in self._ts_col_names.items() if k not in self.timestamp_columns]
        self._original_dt = self._original_dt.drop(old_cols)
        self._ts_col_names = {
            x: f"{x} (ns)" if self._ts_cols[x] == _ALREADY_DT else f"{x} (Local)" for x in self.timestamp_columns
        }
        if self._ts_col_names:
            (foot := self.query_one(TableFooter)).pending_action = "Computing Ts columns"
            try:
                calc_expr = [
                    (
                        pl.col(x).dt.epoch("ns")
                        if self._ts_cols[x] == _ALREADY_DT
                        else pl.from_epoch(pl.col(x), time_unit=self._ts_cols[x]).dt.convert_time_zone(_TIMEZONE)
                    ).alias(self._ts_col_names[x])
                    for x in self.timestamp_columns
                ]
                self._original_dt = self._original_dt.with_columns(calc_expr)
            except Exception as e:
                self.notify(f"Failed to compute timestamp columns: {e}", severity="warn", timeout=5)
            finally:
                foot.pending_action = None
        self.apply_filter()

    @on(SelectFromTable)
    def enable_select_from_table(self, event: SelectFromTable):
        self._select_interest = f"#{event.interested_widget.id}"
        self.query_one("#main_table", CustomTable).focus()

    @on(CustomTable.CellHighlighted, selector="#main_table")
    async def handle_cell_highlight(self, event: CustomTable.CellHighlighted):
        self.cur_row = event.coordinate.row
        self._row_detail.row_df = self._display_dt[self.cur_row]

    @on(CustomTable.CellSelected, selector="#main_table")
    def handle_cell_select(self, event: CustomTable.CellSelected):
        if self._select_interest:
            self.query_one(self._select_interest, ReceivesTableSelect).on_table_select(event.value)
            self._select_interest = None

    @on(Bookmarks.BookmarkSelected)
    def handle_bookmark_select(self, event: Bookmarks.BookmarkSelected):
        dt = self.query_one("#main_table", CustomTable)
        coord = dt.cursor_coordinate
        sel_idx = event.selected_index
        if self.is_filtered:
            filt = self._meta_dt.with_row_index("__displayIndex").filter(pl.col(INDEX_COL) == sel_idx)
            if filt.is_empty():
                self.notify(
                    "Bookmark not present in filtered view.  Remove filters to select this bookmark",
                    severity="error",
                    timeout=5,
                )
                return
            sel_idx = filt["__displayIndex"][0]
        ys = sel_idx
        xs = dt.scroll_x
        dt.scroll_to(xs, ys, animate=False, force=True)
        dt.move_cursor(column=coord.column, row=sel_idx)

    @on(Bookmarks.BookmarkRemoved)
    def handle_bookmark_removed(self, event: Bookmarks.BookmarkRemoved):
        event.stop()
        self.refresh_bindings()
        self.query_one("#main_table", CustomTable).refresh(repaint=True)

    async def action_show_filter(self):
        # import gc
        # gc.set_debug(gc.DEBUG_SAVEALL)
        # def print_gc(*_):
        #     by_typ = {}
        #     for x in gc.garbage:
        #         r = by_typ.setdefault(type(x), 1)
        #         by_typ[type(x)] = r + 1
        #     for x, k in reversed(sorted(by_typ.items(), key=lambda v: v[1])):
        #         print(f"{x}={k}")
        # gc.callbacks.append(print_gc)

        if existing := self.query("#filter"):
            existing.remove()
            return
        self._filter_box.is_goto = False
        await self.mount(self._filter_box, before=self.query_one(TableFooter))

    async def action_show_search(self):
        if existing := self.query("#filter"):
            existing.remove()
            return
        self._filter_box.is_goto = True
        await self.mount(self._filter_box, before=self.query_one(TableFooter))

    def check_action(self, action: str, parameters: tuple[object, ...]) -> bool | None:
        """Check if an action may run."""
        # self.query(".toolbox")
        if not (edtq := self.query_one("#main_table", CustomTable)):
            return False

        if not edtq.has_focus and action in (x.action if isinstance(x, Binding) else x[1] for x in DtBrowser.BINDINGS):
            return False

        match action:
            case "iter_search":
                if not self.active_search_queue:
                    return False
                if bool(parameters[0]) and self.active_search_idx == len(self.active_search_queue) - 1:
                    return False
                if not bool(parameters[0]) and self.active_search_idx == 0:
                    return False
            case "show_bookmarks":
                return self._bookmarks.has_bookmarks
            case "timestamp_selector":
                return len(self._ts_cols) > 0
            case "toggle_row_detail":
                return not self._display_dt.is_empty()

        return True

    @work(exclusive=True)
    async def watch_color_by(self):
        if not self.color_by:
            self._meta_dt = self._meta_dt.drop(COLOR_COL, strict=False)
            self._original_meta = self._original_meta.drop(COLOR_COL, strict=False)
        else:
            (foot := self.query_one(TableFooter)).pending_action = "Recoloring"
            try:
                cols = tuple(sorted(self.color_by))
                if cols not in self._color_by_cache:
                    self._color_by_cache.set(
                        cols,
                        (
                            await self._original_dt.lazy()
                            .with_columns(
                                __color=(
                                    (pl.any_horizontal(*(pl.col(x) != pl.col(x).shift(1) for x in cols)))
                                    .cum_sum()
                                    .fill_null(0)
                                    % len(COLORS.categories)
                                )
                            )
                            .collect_async()
                        )[COLOR_COL],
                    )
                self._original_meta = self._original_meta.with_columns(__color=self._color_by_cache.get(cols))
                self._meta_dt = (
                    await self._meta_dt.lazy()
                    .drop(COLOR_COL, strict=False)
                    .join(self._original_meta.lazy().select([INDEX_COL, COLOR_COL]), how="left", on=INDEX_COL)
                    .collect_async()
                )
            except Exception as e:
                self.notify(f"Failed to apply coloring due to: {e}", severity="error", timeout=10)
            foot.pending_action = None

        self.query_one("#main_table", CustomTable).set_metadata(self._meta_dt)

    def on_mount(self):
        self.cur_total_rows = len(self._display_dt)
        self.total_rows = len(self._original_dt)
        self._row_detail.row_df = self._display_dt[0]
        if self.removed_cols:
            err_str = ", ".join(f"{k}: {v}" for k, v in self.removed_cols.items())
            self.notify(
                f"Removed column(s) with unsupported dtypes: {err_str}",
                severity="warning",
                timeout=10,
            )
        gc.disable()
        # self.set_interval(3, self._maybe_gc)
        # message_hook.set(self._set_last_message)

    def compose(self) -> ComposeResult:
        """Create child widgets for the app."""
        self._ts_col_selector.data_bind(
            selected_columns=DtBrowser.timestamp_columns, available_columns=DtBrowser.available_timestamp_columns
        )
        self._color_selector.data_bind(selected_columns=DtBrowser.color_by, available_columns=DtBrowser.all_columns)
        self._column_selector.data_bind(
            selected_columns=DtBrowser.visible_columns, available_columns=DtBrowser.all_columns
        )

        with Horizontal(id="main_hori"):
            yield TableWithBookmarks(
                self._original_dt,
                metadata_dt=self._meta_dt,
                bookmarks=self._bookmarks,
                cursor_type=CustomTable.CursorType.CELL,
                id="main_table",
            ).data_bind(DtBrowser.active_search_queue)
        yield TableFooter().data_bind(
            DtBrowser.cur_row,
            DtBrowser.cur_total_rows,
            DtBrowser.is_filtered,
            DtBrowser.total_rows,
            DtBrowser.active_search_queue,
            DtBrowser.active_search_idx,
        )


class DtBrowserApp(App):  # pylint: disable=too-many-public-methods,too-many-instance-attributes

    def __init__(self, table_name: str, source_file_or_table: pathlib.Path | pl.DataFrame):
        super().__init__()
        self._table_name = table_name
        self._source = source_file_or_table

    def compose(self):
        yield DtBrowser(self._table_name, self._source)


@click.command()
@click.argument("source_file", nargs=1, required=True, type=pathlib.Path)
def run(source_file: pathlib.Path):
    app = DtBrowserApp(source_file, source_file)
    app.run(mouse=False)


if __name__ == "__main__":
    run()  # pylint: disable=no-value-for-parameter



import pathlib
from dataclasses import dataclass

from textual import on
from textual.binding import Binding
from textual.containers import Horizontal, Vertical
from textual.message import Message
from textual.reactive import reactive
from textual.suggester import Suggester
from textual.widgets import Input, Label, ListItem, ListView, Rule

from dt_browser import HasState, ReceivesTableSelect


class FilterBox(ReceivesTableSelect, HasState):
    DEFAULT_CSS = """
FilterBox {
    height: 15;
    border: tall white;

}

.filterbox--filterrow {
    height: 3;

}

.filterbox--input {
    width: 1fr;
    margin: 0 1;

}

.filterbox--history {
    padding: 0 1;
}
"""

    BINDINGS = [
        ("escape", "close()", "Close"),
        Binding("tab", "toggle_tab()", show=False),
    ]

    @dataclass
    class FilterSubmitted(Message):
        value: str | None

    @dataclass
    class GoToSubmitted(Message):
        value: str | None

    is_goto = reactive(False)

    def __init__(self, *args, suggestor: Suggester | None = None, **kwargs):
        super().__init__(*args, **kwargs)
        self._history: list[str] = []
        self._history_file = pathlib.Path("~/.cache/dtbrowser/filters.txt").expanduser()
        self._history_file.parent.mkdir(exist_ok=True, parents=True)
        if self._history_file.exists():
            with self._history_file.open("r", encoding="utf-8") as f:
                self._history = [x.rstrip() for x in f.readlines()]

        self._active_filter: dict[bool, str | None] = {True: None, False: None}
        self._suggestor = suggestor

    def query_failed(self, query: str):
        self._history.remove(query)
        for child in self.walk_children(ListItem):
            if child.name == query:
                child.remove()

    def on_table_select(self, value: str):
        box = self.query_one(Input)
        box.value = f"{box.value}{value}"
        box.focus()

    @on(Input.Submitted)
    async def apply_filter(self, event: Input.Submitted):
        new_value = event.value
        if new_value == self._active_filter[self.is_goto]:
            return

        the_list = self.query_one(ListView)
        if new_value:
            self.query_one(Input).value = new_value
            if new_value not in self._history[0:10]:
                self._history.append(new_value)
                if len(self._history) < 100:
                    with self._history_file.open("a+", encoding="utf-8") as f:
                        f.write(f"{new_value}\n")
                else:
                    with self._history_file.open("w+", encoding="utf-8") as f:
                        f.writelines([f"{x}\n" for x in self._history[-100:]])

                the_list.index = None
                await the_list.insert(0, [ListItem(Label(new_value), name=new_value)])
                the_list.index = 0
        else:
            the_list.index = None
        self._active_filter[self.is_goto] = new_value
        if self.is_goto:
            msg = FilterBox.GoToSubmitted(value=new_value)
        else:
            msg = FilterBox.FilterSubmitted(value=new_value)
        self.post_message(msg)

    def get_value(self):
        return self._active_filter[self.is_goto]

    @on(ListView.Selected)
    def input_historical(self, event: ListView.Selected):
        box = self.query_one(Input)
        box.value = event.item.name
        box.focus()

    def key_down(self):
        if not (box := self.query_one(ListView)).has_focus and box.children:
            box.focus()

    def action_toggle_tab(self):
        if not (box := self.query_one(Input)).has_focus:
            box.focus()
        if box._suggestion:  # pylint: disable=protected-access
            box.action_cursor_right()

    def action_close(self):
        self.remove()

    def compose(self):
        if self.is_goto:
            self.border_title = "Search dataframe"
        else:
            self.border_title = "Filter dataframe"

        with Vertical():
            with Horizontal(classes="filterbox--filterrow"):
                yield Input(
                    value=self._active_filter[self.is_goto],
                    placeholder="Enter SQL query to " + ("search" if self.is_goto else "filter"),
                    classes="filterbox--input",
                    suggester=self._suggestor,
                )
            yield Rule()
            yield ListView(
                *(ListItem(Label(x), name=x) for x in reversed(self._history)),
                classes="filterbox--history",
            )

    def on_mount(self):
        self.query_one(Input).focus()
        idx = None
        if self._active_filter[self.is_goto] and self._active_filter[self.is_goto] in self._history:
            idx = len(self._history) - (self._history.index(self._active_filter[self.is_goto]) + 1)
        self.query_one(ListView).index = idx



from dataclasses import dataclass

from textual import events, on
from textual.binding import Binding
from textual.message import Message
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import Input, SelectionList
from textual.widgets.selection_list import Selection

_INPUT_HINT = "Type to filter columns list..."


class ColumnSelector(Widget):

    DEFAULT_CSS = """
    ColumnSelector {
        dock: right;
        border: tall $primary;
        height: auto;
    }

    .columnselector--inputbox {
        height: auto;
        margin: 0 0 0 0;
    }
    """

    BINDINGS = [
        ("escape", "close()", "Close and Apply"),
        Binding("ctrl+a", "apply()", "Apply", key_display="ctrl+a"),
        ("shift+up", "move(True)", "Move up"),
        ("shift+down", "move(False)", "Move Down"),
        Binding("ctrl+s", "select_all()", "Select All", key_display="ctrl+s"),
        Binding("ctrl+x", "deselect_all()", "Deselect All", key_display="ctrl+x"),
    ]

    @dataclass
    class ColumnSelectionChanged(Message):
        selected_columns: tuple[str, ...]
        selector: "ColumnSelector"

        @property
        def control(self):
            return self.selector

    available_columns: reactive[tuple[str, ...]] = reactive(tuple())
    selected_columns: reactive[tuple[str, ...]] = reactive(tuple(), init=False)
    ordered_columns: reactive[tuple[str, ...]] = reactive(tuple(), init=False, bindings=True)
    filtered_ordered_columns: reactive[tuple[str, ...]] = reactive(tuple(), init=False, bindings=True)

    def __init__(self, *args, title: str | None = None, allow_reorder: bool = True, **kwargs):
        super().__init__(*args, **kwargs)
        self._allow_reorder = allow_reorder
        self._title = title
        self.filter_value: str | None = None
        self._message: ColumnSelector.ColumnSelectionChanged | None = None

    def action_close(self):
        self.action_apply()
        self.remove()

    def action_apply(self):
        if self._message is not None:
            self.post_message(self._message)
            self._message = None

    def action_deselect_all(self):
        self.query_one(SelectionList).deselect_all()

    def action_select_all(self):
        self.query_one(SelectionList).select_all()

    def check_action(self, action: str, parameters: tuple[object, ...]) -> bool | None:
        if action == "move":
            filter_value = self.query_one(Input).value
            if filter_value:
                return False
            if not self.ordered_columns or not self._allow_reorder:
                return False
            cur_idx = self.query_one(SelectionList).highlighted
            if cur_idx is None:
                return False
            if parameters[0]:
                return cur_idx > 0
            return cur_idx < len(self.ordered_columns) - 1
        return True

    async def on_key(self, event: events.Key) -> None:
        box = self.query_one(Input)
        if box.check_consume_key(event.key, event.character):
            await box.on_event(event)
            return
        if event.key == "backspace":
            box.action_delete_left()
            return

    @on(SelectionList.SelectionHighlighted)
    def _refresh_actions(self):
        self.refresh_bindings()

    @on(Input.Changed)
    def _update_filter(self, event: Input.Changed):
        self.filter_value = event.value
        event.prevent_default()
        self._refresh_options()

    def compute_filtered_ordered_columns(self):
        if not self.filter_value:
            return self.ordered_columns
        return tuple(x for x in self.ordered_columns if self.filter_value.lower() in x.lower())

    def _refresh_options(self):
        sel_list = self.query_one(SelectionList)
        sel_idx = sel_list.highlighted
        if sel_idx is not None:
            sel_val: str | None = sel_list.get_option_at_index(sel_idx).value
        else:
            sel_val = None
        sel_list.clear_options()
        for i, x in enumerate(self.filtered_ordered_columns):
            sel_list.add_option(Selection(x, x, x in self.selected_columns))
            if x == sel_val:
                sel_list.highlighted = i

    def action_move(self, is_up: bool):
        sel_list = self.query_one(SelectionList)
        if (idx := sel_list.highlighted) is None:
            return
        if is_up:
            self.ordered_columns = (
                self.ordered_columns[0 : idx - 1]
                + (self.ordered_columns[idx], self.ordered_columns[idx - 1])
                + self.ordered_columns[idx + 1 :]
            )
        else:
            self.ordered_columns = (
                self.ordered_columns[0:idx]
                + (self.ordered_columns[idx + 1], self.ordered_columns[idx])
                + self.ordered_columns[idx + 2 :]
            )
        self._refresh_options()

    def watch_available_columns(self):
        new_disp = []
        for x in self.available_columns:
            if x not in self.ordered_columns:
                new_disp.append(x)
        if new_disp:
            self.ordered_columns = self.ordered_columns + tuple(new_disp)
        max_item = max([len(x) for x in self.available_columns] + [0])
        self.styles.width = max([max_item + 10, 35, len(_INPUT_HINT)])

    def watch_ordered_columns(self):
        self.selected_columns = [x for x in self.ordered_columns if x in self.selected_columns]
        self._refresh_options()

    def watch_selected_columns(self):
        self._message = ColumnSelector.ColumnSelectionChanged(selected_columns=self.selected_columns, selector=self)

    def on_mount(self):
        self.query_one(SelectionList).focus()
        self.query_one(SelectionList).highlighted = 0

    def compose(self):
        inp = Input(classes="columnselector--inputbox", type="text", placeholder=_INPUT_HINT)
        inp.can_focus = False
        yield inp
        yield SelectionList[int](
            *(Selection(x, x, x in self.selected_columns) for x in self.ordered_columns),
        )

        self.border_title = self._title

    @on(SelectionList.SelectedChanged)
    def on_column_selection(self, event: SelectionList.SelectedChanged):
        event.stop()
        sels = event.selection_list.selected
        new_sels: list[str] = []

        for x in self.ordered_columns:
            # Retain existing selections not in filtered set
            if x not in self.filtered_ordered_columns:
                if x in self.selected_columns:
                    new_sels.append(x)
            elif x in sels:
                new_sels.append(x)

        self.selected_columns = tuple(new_sels)



from textual.suggester import Suggester


class ColumnNameSuggestor(Suggester):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, case_sensitive=True, **kwargs)
        self.columns = tuple[str, ...]()

    async def get_suggestion(self, value: str) -> str | None:
        ori_value = value
        value = value.replace("(", "").replace(")", "")

        if not value:
            return None

        tokens = value.rsplit(maxsplit=3)
        last_token = tokens[-1]
        if len(tokens) > 1:
            combos = ("and", "or", "AND", "OR")
            for misc in combos:
                if misc.startswith(last_token) and tokens[-2] not in combos:
                    return f"{ori_value}{misc.removeprefix(last_token)}"

        for col in self.columns:
            if col.startswith(last_token):
                return f"{ori_value}{col.removeprefix(last_token)}"

        return None
